"""
Generate counterfactual explanation for predictions of tree-based models.
"""
from __future__ import division, print_function

import os

import numpy as np
import tensorflow as tf
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier

from cfxplorer.utils import calculate_distance

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


class Focus:
    """
    FOCUS Lucic, et al. 2022 computes
    Counterfactual Explanations (CFE) using the gradient descent
    method for predictions of the tree-based models.

    Parameters
    ----------
    distance_function: str, optional (default="euclidean")
        Distance function - one of followings;
            - "euclidean"
            - "cosine"
            - "l1"
            - "mahalabobis"

    optimizer: Keras optimizer, optional (default=tf.keras.optimizers.Adam())
        Optimizer for gradient decent

    sigma: float, optional (default=10.0)
        Sigma hyperparameter value for hinge loss

    temperature: float, optional (default=1.0)
        Temperature hyperparameter value for hinge loss

    distance_weight: float, optional (default=0.01)
        Weight hyperparameter for distance loss

    lr: float, optional (default=0.001)
        Learning rate for gradient descent optimization

    num_iter: int, optional (default=100)
        Number of iterations for gradient descent optimization

    direction: str, optional (default="both")
        Direction of perturbation (e.g. both, positive and negative)

    hyperparameter_tuning: bool, optional (default=False)
        if True, generate method returns unchanged_ever and mean_distance

    verbose: int, optional (default=1)
        Verbosity mode.
            - 0: silent
            - else: print current number of iterations

    Reference
    ---------
    Lucic, A., Oosterhuis, H., Haned, H., & de Rijke, M. (2022, June).
    FOCUS: Flexible optimizable counterfactual explanations for tree ensembles.
    In Proceedings of the AAAI Conference on Artificial Intelligence (Vol. 36, No. 5, pp. 5313-5322).

    Examples
    --------
    - Initialize FOCUS on default parameters
    - Generate counterfactual explanations

    focus = Focus()

    cfe_features = focus.generate(model, X)
    """

    def __init__(
        self,
        distance_function="euclidean",
        optimizer=tf.keras.optimizers.Adam(),
        sigma=10.0,
        temperature=1.0,
        distance_weight=0.01,
        lr=0.001,
        num_iter=100,
        direction="both",
        hyperparameter_tuning=False,
        verbose=1,
    ):
        self.distance_function = distance_function
        self.optimizer = optimizer
        self.sigma = sigma
        self.temperature = temperature
        self.distance_weight = distance_weight
        self.lr = lr
        self.num_iter = num_iter
        self.direction = direction
        self.hyperparameter_tuning = hyperparameter_tuning
        self.verbose = verbose

    def generate(self, model, X, x_train=None):
        """
        Generate counterfactual explanations for the
        predictions from a tree-based model.

        Args:
        model: model object
            The machine learning model
                - DecisionTreeClassifier
                - RandomForestClassifier
                - AdaBoostClassifier

        X: numpy array
            The input feature to generate CFE

        x_train: numpy array, optional (default=None)
            The training data features
                - This will be used to calculate Mahalanobis distances

        Returns:
            The best perturbed features

        This method generates counterfactual explanations for the
        predictions made by a tree-based model.
        It uses the gradient descent method to optimize the input features
        based on a combination of hinge loss, approximate probability and a distance term.
        The `model` should be an instance of a tree-based model,
        such as DecisionTreeClassifier, RandomForestClassifier or AdaBoostClassifier.
        The `X` parameter represents the input features for which
        counterfactual explanations are desired.
        The `x_train` parameter is an optional argument that
        represents the training data features used
        to compute the approximate probability.

        The method returns the best perturbed features,
        which represent the optimized input features that
        result in counterfactual explanations.
        """
        X = Focus.prepare_features_by_perturb_direction(model, X, self.direction)

        perturbed = tf.Variable(
            initial_value=X,
            trainable=True,
            name="perturbed_features",
        )
        n_rows = len(X)
        distance_weight: np.ndarray = np.full(n_rows, self.distance_weight)
        to_optimize = [perturbed]
        mask_vector = np.ones(n_rows)
        best_perturb = np.zeros(perturbed.shape)
        best_distance = np.full(n_rows, np.inf)
        predictions = tf.constant(model.predict(X))

        for i in range(1, self.num_iter + 1):
            if self.verbose != 0:
                print(f"iteration {i}")

            grad = Focus.compute_gradient(
                model,
                X,
                predictions,
                to_optimize,
                mask_vector,
                perturbed,
                distance_weight,
                x_train,
                self.distance_function,
                self.sigma,
                self.temperature,
                self.optimizer,
            )

            self.optimizer.apply_gradients(
                zip(grad, to_optimize),
            )
            perturbed.assign(tf.clip_by_value(perturbed, 0, 1))

            distance = calculate_distance(self.distance_function, perturbed, X, x_train)

            cur_predicts = model.predict(perturbed.numpy())
            mask_vector = np.equal(predictions, cur_predicts).astype(np.float32)
            mask_flipped = predictions != cur_predicts

            distance_np = distance.numpy()
            mask_smaller_dist = distance_np < best_distance

            temp_dist = best_distance.copy()
            temp_dist[mask_flipped] = distance_np[mask_flipped]
            best_distance[mask_smaller_dist] = temp_dist[mask_smaller_dist]

            temp_perturb = best_perturb.copy()
            temp_perturb[mask_flipped] = perturbed[mask_flipped]
            best_perturb[mask_smaller_dist] = temp_perturb[mask_smaller_dist]

        if self.hyperparameter_tuning:
            return (
                best_perturb,
                len(best_distance[best_distance == np.inf]),
                np.mean(best_distance[best_distance != np.inf]),
            )

        else:
            print(
                f"The number of rows that are unchanged ever is "
                f"{len(best_distance[best_distance == np.inf])}"
            )
            print(
                f"The mean distance is "
                f"{np.mean(best_distance[best_distance != np.inf])}"
            )

            return best_perturb

    @staticmethod
    def prepare_features_by_perturb_direction(model, X: np.ndarray, direction: str):
        """
        Prepares the input data `X` based on the perturbation direction.

        Args:
            model (object): The model used for predicting the labels.
            X (np.ndarray): The input data to be prepared based on the perturbation direction.
            direction (str): The perturbation direction to consider.
                                Available options: "positive", "negative", "both".

        Returns:
            np.ndarray: The prepared input data based on the perturbation direction.

        Raises:
            ValueError: If an invalid `direction` is provided.

        This method filters and prepares the input data `X` based on the perturbation direction
        specified by the `direction` argument.
        It uses the provided `model` to predict the labels for the input data.
        The available options for `direction` are:
        - "positive":
            Returns the subset of input data where the model predicts the label as 0.
        - "negative":
            Returns the subset of input data where the model predicts the label as 1.
        - "both":
            Returns the input data as is without any filtering.

        Note that the `model` object should have a `predict` method
        that returns the predicted labels.
        """
        if direction == "positive":
            return X[model.predict(X) == 0]
        elif direction == "negative":
            return X[model.predict(X) == 1]
        elif direction == "both":
            return X
        else:
            raise ValueError(f"direction {direction} is not available")

    @staticmethod
    def compute_gradient(
        model,
        X,
        predictions,
        to_optimize,
        mask_vector,
        perturbed,
        distance_weight,
        x_train,
        distance_function,
        sigma,
        temperature,
        optimizer,
    ):
        """
        Computes the gradient of the loss function with respect to the variables to optimize.

        Returns:
        tf.Tensor: The computed gradient of the loss function with respect to the variables to optimize.

        This method computes the gradient of the loss function based on the provided inputs.
        It uses a TensorFlow GradientTape to record the operations for automatic differentiation.
        The loss function is defined as a combination of hinge loss, approximate probability, and a distance term.
        The gradient is then calculated with respect to the variables specified in the `to_optimize` list.
        """
        prediction_class_index = tf.stack(
            (tf.constant(np.arange(len(X), dtype=int)), predictions), axis=1
        )

        with tf.GradientTape(persistent=True) as tape:
            hinge_loss = Focus.filter_hinge_loss(
                len(model.classes_),
                mask_vector,
                perturbed,
                sigma,
                temperature,
                model,
            )
            approx_prob = tf.gather_nd(hinge_loss, prediction_class_index)
            distance = calculate_distance(distance_function, perturbed, X, x_train)
            hinge_approx_prob = tf.cast(mask_vector * approx_prob, tf.float32)
            loss = tf.reduce_mean(
                hinge_approx_prob + distance_weight * tf.cast(distance, tf.float32)
            )

        return tape.gradient(loss, to_optimize)

    @staticmethod
    def parse_class_tree(tree, X, sigma: float) -> list:
        """
        Compute impurity of each leaf node in a decision tree and approximate it using sigmoid function.

        Args:
        tree (DecisionTreeClassifier): Trained decision tree model.
        X (np.ndarray): Input feature values.
        sigma (float): Scaling factor to apply to sigmoid activation.

        Returns:
        list: A list of impurity values for each class label in the tree.
        """
        n_nodes = tree.tree_.node_count
        children_left = tree.tree_.children_left
        children_right = tree.tree_.children_right
        feature = tree.tree_.feature
        threshold = tree.tree_.threshold
        values = tree.tree_.value

        nodes = [None] * n_nodes
        leaf_nodes = [[] for _ in range(len(tree.classes_))]

        node_depth = np.zeros(shape=n_nodes, dtype=np.int32)
        is_leaves = np.zeros(shape=n_nodes, dtype=bool)
        stack = [(0, 0)]

        while len(stack) > 0:
            node_id, depth = stack.pop()
            node_depth[node_id] = depth

            is_split_node = children_left[node_id] != children_right[node_id]
            if is_split_node:
                stack.append((children_left[node_id], depth + 1))
                stack.append((children_right[node_id], depth + 1))
            else:
                is_leaves[node_id] = True

        for i in range(n_nodes):
            cur_node = nodes[i]
            if children_left[i] != children_right[i]:
                if cur_node is None:
                    cur_node = 1.0

                sigma = np.full(len(X), sigma)
                activation = tf.math.sigmoid((X[:, feature[i]] - threshold[i]) * sigma)

                left_node, right_node = 1.0 - activation, activation
                nodes[children_left[i]], nodes[children_right[i]] = (
                    cur_node * left_node,
                    cur_node * right_node,
                )

            else:
                max_class = np.argmax(values[i])
                leaf_nodes[max_class].append(cur_node)

        return leaf_nodes

    @staticmethod
    def get_prob_classification_tree(tree, X, sigma: float) -> tf.Tensor:
        """
        get_prob_classification_tree - computes the probability of each sample's classification in a decision tree

        Args:
        tree (DecisionTreeClassifier): a fitted decision tree model
        X (tf.Tensor): a tensor of input features
        sigma (float): parameter for Gaussian smoothing
        Outputs:

        Returns:
        prob_stacked (tf.Tensor): a tensor of probabilities for each sample's classification
        This function returns the probabilities of each sample's classification in a decision tree.
        It calculates the impurities of the leaf nodes that each sample falls into, and then
        computes the sum of these impurities for each class. If the tree has only one node,
        the probability of the correct class is set to 1 and the probability of the incorrect class is set to 0.
        The final result is a tensor of stacked probabilities for each sample's classification.
        """

        leaf_nodes = Focus.parse_class_tree(tree, X, sigma)

        if tree.tree_.node_count > 1:
            prob_list = [sum(leaf_nodes[c_i]) for c_i in range(len(tree.classes_))]
            i = 0
            while i < len(prob_list):
                if prob_list[i].numpy().all() == 0:
                    prob_list.pop(i)
                else:
                    i += 1

            prob_stacked = tf.stack(prob_list, axis=-1)

        else:  # sometimes tree only has one node
            only_class = tree.predict(
                tf.reshape(X[0, :], shape=(1, -1))
            )  # can differ depending on particular samples used to train each tree

            correct_class = tf.constant(
                1, shape=(len(X)), dtype=tf.float32
            )  # prob(belong to correct class) = 100 since there's only one node
            incorrect_class = tf.constant(
                0, shape=(len(X)), dtype=tf.float32
            )  # prob(wrong class) = 0
            if only_class == 1.0:
                class_0 = incorrect_class
                class_1 = correct_class
            elif only_class == 0.0:
                class_0 = correct_class
                class_1 = incorrect_class
            else:
                raise ValueError("The class should be either 0 or 1")
            class_labels = [class_0, class_1]
            prob_stacked = tf.stack(class_labels, axis=1)
        return prob_stacked

    @staticmethod
    def get_prob_classification_forest(
        model, X: tf.Tensor, sigma: float, temperature: float
    ) -> tf.Tensor:
        """
        Calculate the softmax probabilities for classification for a random forest or AdaBoost model.

        Args:
        model: RandomForestClassifier or AdaBoostClassifier
            The trained decision tree model.
        X: tf.Tensor
            The input feature matrix.
        sigma: float
            The sigma value used to compute the activation of nodes in each decision tree.
        temperature: float
            The temperature to adjust the scale of the logits.

        Returns:
        tf.Tensor
            The softmax probabilities of the classification.
        """
        dt_prob_list = [
            Focus.get_prob_classification_tree(estimator, X, sigma)
            for estimator in model.estimators_
        ][:100]

        if isinstance(model, AdaBoostClassifier):
            weights = model.estimator_weights_
        elif isinstance(model, RandomForestClassifier):
            weights = np.full(len(model.estimators_), 1 / len(model.estimators_))
        else:
            raise ValueError(
                "model object should be either AdaBoostClassifier or RandomForestClassifier"
            )

        logits = sum(weight * tree for weight, tree in zip(weights, dt_prob_list))

        temperature = np.full(len(X), temperature)
        if type(temperature) in [float, int]:
            expits = tf.exp(temperature * logits)
        else:
            expits = tf.exp(temperature[:, None] * logits)

        softmax = expits / tf.reduce_sum(expits, axis=1)[:, None]

        return softmax

    @staticmethod
    def filter_hinge_loss(
        n_class,
        mask_vector,
        X,
        sigma,
        temperature,
        model,
    ) -> tf.Tensor:
        """
        Calculates the filtered probabilities of each data point for the given model.

        Args:
        n_class (int): Number of classes.
        mask_vector (np.ndarray): A boolean mask indicating which data points should be considered.
        X (tf.Tensor): The feature input for the model.
        sigma (float): The value of sigma for computing the probabilities.
        temperature (float): The temperature to be used for the softmax function.
        model: The machine learning model;
            e.g., DecisionTreeClassifier, RandomForestClassifier, AdaBoostClassifier.

        Returns:
        hinge_loss (tf.Tensor): The filtered probabilities of each data point.
        """
        n_input = X.shape[0]

        filtered_input = tf.boolean_mask(X, mask_vector)

        if not isinstance(model, DecisionTreeClassifier):
            filtered_loss = Focus.get_prob_classification_forest(
                model, filtered_input, sigma=sigma, temperature=temperature
            )
        elif isinstance(model, DecisionTreeClassifier):
            filtered_loss = Focus.get_prob_classification_tree(
                model, filtered_input, sigma
            )

        indices = np.where(mask_vector)[0]
        hinge_loss = tf.tensor_scatter_nd_add(
            np.zeros((n_input, n_class)),
            indices[:, None],
            filtered_loss,
        )
        return hinge_loss



import pytest
import tensorflow as tf
from sklearn.datasets import make_classification
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier

from cfxplorer import Focus

x_train, y_train = make_classification(
    n_samples=200, n_features=10, n_classes=2, random_state=42
)
x_test, y_test = make_classification(
    n_samples=100, n_features=10, n_classes=2, random_state=42
)

dt_model = DecisionTreeClassifier(random_state=42)
dt_model.fit(x_train, y_train)

ab_model = AdaBoostClassifier(random_state=42)
ab_model.fit(x_train, y_train)

rf_model = RandomForestClassifier(random_state=42)
rf_model.fit(x_train, y_train)

model_data = [
    (dt_model, x_test),
    (ab_model, x_test),
    (rf_model, x_test),
]

focus_model_data = [
    (dt_model, x_test, None, tf.keras.optimizers.Adam(), "euclidean", False),
    (ab_model, x_test, None, tf.keras.optimizers.Adam(), "l1", False),
    (rf_model, x_test, None, tf.keras.optimizers.Adam(), "euclidean", False),
    (dt_model, x_test, x_train, tf.keras.optimizers.Adam(), "mahalanobis", False),
    (ab_model, x_test, x_train, tf.keras.optimizers.Adam(), "mahalanobis", False),
    (rf_model, x_test, x_train, tf.keras.optimizers.Adam(), "mahalanobis", False),
    (dt_model, x_test, None, tf.keras.optimizers.Adam(), "cosine", True),
    (ab_model, x_test, None, tf.keras.optimizers.Adam(), "cosine", True),
    (rf_model, x_test, None, tf.keras.optimizers.Adam(), "l1", True),
    (dt_model, x_test, x_train, tf.keras.optimizers.Adam(), "mahalanobis", True),
    (ab_model, x_test, x_train, tf.keras.optimizers.Adam(), "mahalanobis", True),
    (rf_model, x_test, x_train, tf.keras.optimizers.Adam(), "mahalanobis", True),
    (dt_model, x_test, None, tf.keras.optimizers.Nadam(), "euclidean", False),
    (ab_model, x_test, None, tf.keras.optimizers.RMSprop(), "l1", False),
    (rf_model, x_test, None, tf.keras.optimizers.Ftrl(), "euclidean", False),
    (dt_model, x_test, x_train, tf.keras.optimizers.Ftrl(), "mahalanobis", False),
    (ab_model, x_test, x_train, tf.keras.optimizers.Nadam(), "mahalanobis", False),
    (rf_model, x_test, x_train, tf.keras.optimizers.SGD(), "mahalanobis", False),
    (dt_model, x_test, None, tf.keras.optimizers.SGD(), "cosine", True),
    (ab_model, x_test, None, tf.keras.optimizers.Adagrad(), "cosine", True),
    (rf_model, x_test, None, tf.keras.optimizers.Adadelta(), "l1", True),
    (dt_model, x_test, x_train, tf.keras.optimizers.Adamax(), "mahalanobis", True),
    (ab_model, x_test, x_train, tf.keras.optimizers.Adamax(), "mahalanobis", True),
    (rf_model, x_test, x_train, tf.keras.optimizers.Adadelta(), "mahalanobis", True),
]


@pytest.mark.parametrize("model, X", model_data)
def test_prepare_features_by_perturb_direction(model, X):
    focus = Focus()
    direction = "positive"
    prepared_x = focus.prepare_features_by_perturb_direction(model, X, direction)
    assert len(prepared_x) < len(X)


@pytest.mark.parametrize(
    "model, x_test, x_train, optimizer, distance_function, hyperparameter_tuning",
    focus_model_data,
)
def test_generate(
    model, x_test, x_train, optimizer, distance_function, hyperparameter_tuning
):
    """Test `generate` method by using multiple combinations of different parameters"""
    focus = Focus(
        num_iter=2,
        optimizer=optimizer,
        distance_function=distance_function,
        hyperparameter_tuning=hyperparameter_tuning,
    )
    if hyperparameter_tuning:
        best_perturb, unchanged_ever, best_distance = focus.generate(
            model, x_test, x_train
        )

        assert best_perturb.all() != x_test.all()
        assert isinstance(unchanged_ever, int)
        assert isinstance(best_distance, float)

    else:
        best_perturb = focus.generate(model, x_test, x_train)
        assert best_perturb.all() != x_test.all()



import numpy as np
import tensorflow as tf


def safe_euclidean(matrix_diff, epsilon=10.0**-10) -> tf.Tensor:
    """
    Calculates the Euclidean distance between two matrices with a small epsilon added to prevent singularities.

    Args:
    matrix_diff: A tensor representing the difference between two matrices
    epsilon (float): A small number added to prevent singularities in the calculation (default 10.0 ** -10)

    Returns:
    tf.Tensor: A tensor representing the Euclidean distance between the two matrices
    """
    return (tf.reduce_sum(matrix_diff**2, axis=-1) + epsilon) ** 0.5


def safe_cosine(feat_input, perturbed, epsilon=10.0**-10) -> tf.Tensor:
    """
    Calculates cosine distance between two input arrays `feat_input` and `perturbed`
    while ensuring numerical stability with `epsilon`.

    Args:
        feat_input: The first input array to calculate cosine distance with.
        perturbed: The second input array to calculate cosine distance with.
        epsilon (float, optional): A small value added to the denominator to prevent division by zero.
        Defaults to 10 ** -10.

    Returns:
        tf.Tensor: The cosine distance between `feat_input` and `perturbed` as a tensor.
    """
    normalize_x1 = tf.nn.l2_normalize(feat_input)
    normalize_x2 = tf.nn.l2_normalize(perturbed)
    cosine_loss = tf.keras.losses.CosineSimilarity(
        axis=-1,
        reduction=tf.keras.losses.Reduction.NONE,
    )
    dist = 1 - cosine_loss(normalize_x1, normalize_x2) + epsilon

    dist = tf.cast(tf.squeeze(dist), tf.float32)
    return dist


def safe_l1(matrix_diff, epsilon=10.0**-10) -> tf.Tensor:
    """
    Calculates the L1 (Manhattan) distance between two tensors with a small epsilon value
    added to prevent division by zero.

    Args:
    matrix_diff: Tensor to calculate L1 distance.
    epsilon: A small value to avoid division by zero (default: 10.0 ** -10).

    Returns:
    The L1 distance between the input tensors with the epsilon value added.
    """
    return tf.reduce_sum(tf.abs(matrix_diff), axis=1) + epsilon


def tf_cov(x_train) -> tf.Tensor:
    """
    Computes the covariance matrix of the input feature matrix x_train.

    Args:
    x_train: the train set

    Returns:
    cov_xx: a TensorFlow tensor of shape (num_features, num_features) representing the covariance matrix.
    """
    mean_x = tf.reduce_mean(x_train, axis=0, keepdims=True)
    mx = tf.matmul(tf.transpose(mean_x), mean_x)
    vx = tf.matmul(tf.transpose(x_train), x_train) / tf.cast(
        tf.shape(x_train)[0], tf.float64
    )
    cov_xx = vx - mx
    return cov_xx


def safe_mahal(matrix_diff, x_train, epsilon=10.0**-10) -> tf.Tensor:
    """
    Calculates Mahalanobis distance using TensorFlow

    Args:
    matrix_diff: A tensor of shape (N, D) representing the difference between perturbed and feat_input
    x_train: The training data.
    epsilon: A scalar value to be added to the diagonal of covariance matrix
    to make it invertible.
    Returns:

    A tensor of shape (N,) representing the Mahalanobis distance between each sample
    and the mean of the training data.
    """
    covar = tf_cov(x_train)
    inv_covar = tf.linalg.inv(covar)
    return tf.reduce_sum(
        tf.multiply(tf.matmul(matrix_diff + epsilon, inv_covar), matrix_diff + epsilon),
        axis=1,
    )


def calculate_distance(
    distance_function: str,
    perturbed: tf.Variable,
    feat_input: np.ndarray,
    x_train: np.ndarray = None,
) -> tf.Tensor:
    """
    Calculates the distance between the perturbed and feat_input data using the specified distance function.

    Args:
    - distance_function (str): The distance function to use,
        can be either "euclidean", "cosine", "l1" or "mahalanobis".
    - perturbed (tf.Variable): The perturbed data.
    - feat_input (np.ndarray): The original feature data.
    - x_train (np.ndarray, optional): The training data, required for the Mahalanobis distance calculation.
    Defaults to None.

    Returns:
    - tf.Tensor: The calculated distance.

    Raises:
    - ValueError: If `distance_function` is "mahal" but `x_train` is not provided.
    """
    if distance_function == "euclidean":
        return safe_euclidean(perturbed - feat_input)
    elif distance_function == "cosine":
        return safe_cosine(feat_input, perturbed)
    elif distance_function == "l1":
        return safe_l1(perturbed - feat_input)
    elif distance_function == "mahalanobis":
        try:
            return safe_mahal(perturbed - feat_input, x_train)
        except ValueError:
            raise ValueError("x_train is empty")



import random

import numpy as np
import pytest
import tensorflow as tf

from cfxplorer.utils import (
    calculate_distance,
    safe_cosine,
    safe_euclidean,
    safe_l1,
    safe_mahal,
    tf_cov,
)

epsilon = 10.0**-10
random.seed(42)

covariance_test_data = [
    (
        tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=tf.float64),
        tf.constant(
            [[6.0, 6.0, 6.0], [6.0, 6.0, 6.0], [6.0, 6.0, 6.0]], dtype=tf.float64
        ),
    ),
    (
        tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=tf.float64),
        tf.convert_to_tensor(
            np.cov(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).T, bias=True),
            dtype=tf.float64,
        ),
    ),
]

distance_test_data = [
    (
        np.array(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 2, 3], [1, 2, 3]], dtype=np.float64
        ),
        np.array(
            [[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 2, 3], [1, 2, 3]], dtype=np.float64
        ),
    )
]

calculate_distance_data = [
    # Euclidean
    (
        "euclidean",
        tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=tf.float64),
        np.array([[4, 5, 6], [7, 8, 9], [1, 2, 3]]),
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
        np.array([5.1961524227162545, 5.1961524227162545, 10.392304845418074]),
    ),
    # Cosine
    (
        "cosine",
        tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=tf.float64),
        np.array([[4, 5, 6], [7, 8, 9], [1, 2, 3]], dtype=np.float64),
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.float64),
        np.array([1.9746319, 1.9981909, 1.959412]),
    ),
    # Manhattan
    (
        "l1",
        tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=tf.float64),
        np.array([[4, 5, 6], [7, 8, 9], [1, 2, 3]]),
        np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
        np.array([9, 9, 18]),
    ),
    # Mahalanobis
    (
        "mahalanobis",
        tf.constant([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=tf.float64),
        np.array([[1, 2, 0], [6, 4, 5], [9, 7, 8]], dtype=np.float64),
        np.array([[4.9, 9.9, 4.1], [7.8, 9.9, 1.4], [2.9, 9.9, 3.2]]),
        np.array([-1.20098143e08, -1.20019050e28, -1.20019050e28]),
    ),
]


@pytest.mark.parametrize("feat_input_cov, expected_output_cov", covariance_test_data)
def test_tf_cov(feat_input_cov, expected_output_cov):
    assert tf_cov(feat_input_cov).numpy() == pytest.approx(
        expected_output_cov.numpy(), rel=1e-5, abs=1e-5
    )


@pytest.mark.parametrize("feat_input, feat_input2", distance_test_data)
def test_safe_euclidean(feat_input, feat_input2):
    expected = (np.sum(feat_input2**2, axis=-1) + epsilon) ** 0.5
    assert safe_euclidean(feat_input).numpy() == pytest.approx(expected)


@pytest.mark.parametrize("feat_input, feat_input2", distance_test_data)
def test_safe_cosine(feat_input, feat_input2):
    cosine_distance = safe_cosine(feat_input, feat_input, epsilon)

    # Assert that the shape of the output is as expected
    assert cosine_distance.shape == (len(feat_input),)


@pytest.mark.parametrize("feat_input, feat_input2", distance_test_data)
def test_safe_l1(feat_input, feat_input2):
    expected = np.sum(abs(feat_input2), axis=1) + epsilon
    assert safe_l1(feat_input).numpy() == pytest.approx(expected)


@pytest.mark.parametrize("feat_input, feat_input2", distance_test_data)
def test_safe_mahal(feat_input, feat_input2):
    """
    This unit test checks the function's accuracy in producing the expected output values.
    The correctness of the function's underlying logic has been previously verified.
    """
    mahalanobis_distance = safe_mahal(feat_input, feat_input2)

    # Assert that the shape of the output is as expected
    assert mahalanobis_distance.shape == (len(feat_input),)


@pytest.mark.parametrize(
    "distance_function, perturbed, feat_input, x_train, expected_output",
    calculate_distance_data,
)
def test_calculate_distance(
    distance_function, perturbed, feat_input, x_train, expected_output
):
    assert calculate_distance(
        distance_function, perturbed, feat_input, x_train
    ).numpy() == pytest.approx(expected_output)



from cfxplorer.focus import Focus

# Let users know if they're missing any of our hard dependencies
_hard_dependencies = ("tensorflow", "numpy")
_missing_dependencies = []

for _dependency in _hard_dependencies:
    try:
        __import__(_dependency)
    except ImportError as _e:  # pragma: no cover
        _missing_dependencies.append(f"{_dependency}: {_e}")

if _missing_dependencies:  # pragma: no cover
    raise ImportError(
        "Unable to import required dependencies:\n" + "\n".join(_missing_dependencies)
    )
del _hard_dependencies, _dependency, _missing_dependencies

__all__ = ["Focus"]



from io import open
from os import path

from setuptools import find_packages, setup

DESCRIPTION = "CFXplorer is a python package for generating counterfactual explanations for given model and feature set"

# get __version__ from _version.py
ver_file = path.join("cfxplorer", "version.py")
with open(ver_file) as f:
    exec(f.read())

this_directory = path.abspath(path.dirname(__file__))


# read the contents of README.rst
def readme():
    with open(path.join(this_directory, "README.rst"), encoding="utf-8") as f:
        return f.read()


setup(
    name="CFXplorer",
    version=__version__,
    author="Kyosuke Morita",
    author_email="kq441morita@gmail.com",
    description=DESCRIPTION,
    long_description=readme(),
    packages=find_packages(),
    install_requires=[],
    keywords=[
        "python",
        "counterfactual explanation",
        "binary classification",
        "machine learning",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)
